/*
 * Copyright (c) 2025 DuckDuckGo
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.duckduckgo.app.browser.navigation.bar

import android.view.ViewTreeObserver
import androidx.constraintlayout.widget.ConstraintSet
import androidx.core.view.marginBottom
import com.duckduckgo.app.browser.BrowserTabFragment
import com.duckduckgo.app.browser.databinding.FragmentBrowserTabBinding
import com.duckduckgo.app.browser.navigation.bar.view.BrowserNavigationBarObserver
import com.duckduckgo.app.browser.navigation.bar.view.BrowserNavigationBarView
import com.duckduckgo.app.browser.omnibar.Omnibar
import com.duckduckgo.app.browser.omnibar.model.OmnibarPosition.BOTTOM
import com.duckduckgo.app.browser.omnibar.model.OmnibarPosition.TOP
import com.duckduckgo.app.browser.omnibar.model.OmnibarType.FADE
import com.duckduckgo.app.browser.omnibar.model.OmnibarType.SCROLLING
import com.duckduckgo.app.browser.viewstate.BrowserViewState
import com.duckduckgo.common.ui.experiments.visual.store.VisualDesignExperimentDataStore
import com.duckduckgo.common.ui.view.gone
import com.duckduckgo.common.ui.view.show
import com.duckduckgo.common.utils.keyboardVisibilityFlow
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch

/**
 * Helper class that extracts business logic that manages the [BrowserNavigationBarView] from the [BrowserTabFragment].
 *
 * The class needs to be instantiated strictly after the fragment's view has been created,
 * and [onDestroyView] has to be called when the the fragment's view is destroyed.
 * After the view is destroyed, the class becomes no-op and needs to be re-instantiated with a valid view binding.
 */
class BrowserNavigationBarViewIntegration(
    private val lifecycleScope: CoroutineScope,
    private val browserTabFragmentBinding: FragmentBrowserTabBinding,
    private val visualDesignExperimentDataStore: VisualDesignExperimentDataStore,
    private val omnibar: Omnibar,
    browserNavigationBarObserver: BrowserNavigationBarObserver,
) {

    private val navigationBarView: BrowserNavigationBarView = browserTabFragmentBinding.navigationBar

    private var stateObserverJob: Job? = null
    private var keyboardVisibilityJob: Job? = null

    private val topFadeOmnibarOffsetChangedListener = ViewTreeObserver.OnPreDrawListener {
        val ratio = browserTabFragmentBinding.fadeOmnibar.getShiftRatio()
        updateNavigationBarMargin(-ratio)
        true
    }

    private val bottomFadeOmnibarPreDrawListener = ViewTreeObserver.OnPreDrawListener {
        val ratio = browserTabFragmentBinding.fadeOmnibarBottom.getShiftRatio()
        updateNavigationBarMargin(-ratio)
        true
    }

    init {
        stateObserverJob = lifecycleScope.launch {
            visualDesignExperimentDataStore.navigationBarState.collect { navigationBarState ->
                if (navigationBarState.isEnabled) {
                    onEnabled()
                } else {
                    onDisabled()
                }
            }
        }
        navigationBarView.browserNavigationBarObserver = browserNavigationBarObserver
    }

    fun renderBrowserViewState(viewState: BrowserViewState) {
        navigationBarView.setCanGoBack(viewState.canGoBack)
        navigationBarView.setCanGoForward(viewState.canGoForward)
    }

    fun configureCustomTab() {
        navigationBarView.setCustomTab(isCustomTab = true)
    }

    fun onDestroyView() {
        stateObserverJob?.cancel()
        onDisabled()
    }

    private fun onEnabled() {
        // we're hiding the navigation bar when keyboard is shown,
        // to prevent it from being "pushed up" within the coordinator layout
        keyboardVisibilityJob = lifecycleScope.launch {
            omnibar.textInputRootView.keyboardVisibilityFlow().collect { keyboardVisible ->
                if (keyboardVisible) {
                    navigationBarView.gone()
                } else {
                    navigationBarView.show()
                }
            }
        }

        when (omnibar.omnibarPosition) {
            TOP -> {
                when (omnibar.omnibarType) {
                    SCROLLING -> {
                        // no-op
                    }

                    FADE -> {
                        browserTabFragmentBinding.fadeOmnibar.viewTreeObserver.addOnPreDrawListener(topFadeOmnibarOffsetChangedListener)
                    }
                }
            }

            BOTTOM -> {
                when (omnibar.omnibarType) {
                    SCROLLING -> {
                        // no-op
                    }

                    FADE -> {
                        browserTabFragmentBinding.fadeOmnibarBottom.viewTreeObserver.addOnPreDrawListener(bottomFadeOmnibarPreDrawListener)
                    }
                }
            }
        }
    }

    private fun onDisabled() {
        keyboardVisibilityJob?.cancel()
        browserTabFragmentBinding.fadeOmnibar.viewTreeObserver.removeOnPreDrawListener(topFadeOmnibarOffsetChangedListener)
        browserTabFragmentBinding.fadeOmnibarBottom.viewTreeObserver.removeOnPreDrawListener(bottomFadeOmnibarPreDrawListener)
    }

    private fun updateNavigationBarMargin(offsetPercentage: Float) {
        if (navigationBarView.isAttachedToWindow) {
            val margin = (navigationBarView.measuredHeight * offsetPercentage).toInt()
            if (margin != navigationBarView.marginBottom) {
                val constraintLayout = browserTabFragmentBinding.rootContainer
                val constraintSet = ConstraintSet()
                constraintSet.clone(constraintLayout)
                constraintSet.setMargin(navigationBarView.id, ConstraintSet.BOTTOM, margin)
                constraintSet.connect(browserTabFragmentBinding.rootContainer.id, ConstraintSet.BOTTOM, navigationBarView.id, ConstraintSet.TOP, 0)
                constraintSet.applyTo(constraintLayout)
            }
        }
    }
}
